@@ -6,14 +6,6 @@
 
 package config
 
-type FolderType int
-
-const (
-	FolderTypeSendReceive FolderType = iota // default is sendreceive
-	FolderTypeSendOnly
-	FolderTypeReceiveOnly
-)
-
 func (t FolderType) String() string {
 	switch t {
 	case FolderTypeSendReceive:
